'''
Created on Mar 7, 2016

@author: mike
'''
from patgen import EMPTYSET
from patgen.suffix_array import SuffixArray


class PatternSet(list):

    def __init__(self):
        list.__init__(self)

    @property
    def maxchunk(self):
        if len(self) == 0:
            return 0
        
        return max(x.maxchunk for x in self)

    def hyphenate(self, word, margins):
        
        prediction = set()
        for i, layer in enumerate(self):
            if (i & 1) == 0:
                # hyphenation layer
                prediction.update(layer.predict(word, margins))
            else:
                # inhibiting layer
                prediction.difference_update(layer.predict(word, margins))
        
        return prediction
    
    def keys(self):
        keys = set()
        for layer in self:
            keys.update(layer.keys())
        return keys

    def pattern_strings(self):

        for key in sorted(self.keys()):
            control = self.get_pattern_control(key)
            
            if control:
                yield self.format_pattern(key, control)

    def get_pattern_control(self, key):

        control = {}
        for level0, layer in enumerate(self):
            for index in layer.get(key, EMPTYSET):
                control[index] = level0 + 1

        return control

    def set_pattern_control(self, key, control):
        for layer in self:
            layer[key].clear()
            
        for i, level in control.items():
            if level > 0:  # paranoid
                layer = self[level - 1]
                layer[key].add(i)

    def compact(self):
        patterns = {}

        for key in self.keys():
            patterns[key] = self.get_pattern_control(key)

        suffix_array = SuffixArray.build(patterns.keys())

        # patterns that are substrings of other patterns
        # can cancel longer pattern's control
        for key, control in patterns.items():
            for lkey, offset in suffix_array.superstrings(key):
                if lkey == key:
                    continue  # do not attempt to compact self!

                lcontrol = patterns[lkey]
                for i, val in control.items():
                    lval = lcontrol.get(offset + i, 0)
                    if 0 < lval <= val:
                        del lcontrol[offset + i] 

        for key, control in patterns.items():
            self.set_pattern_control(key, control)

    @staticmethod
    def format_pattern(key, control):
        out = []
        for i in range(len(key) + 1):
            if i > 0:
                out.append(key[i-1])
            c = control.get(i, 0)
            if c > 0:
                out.append(str(c))
        return ''.join(out)

    def errors(self, dictionary, margins):

        for word, hyphens in dictionary.items():
            prediction = self.hyphenate(word, margins=margins) 

            missed = hyphens - prediction
            false  = prediction - hyphens
            
            if missed or false:
                yield word, hyphens, missed, false

    def evaluate(self, dictionary, margins):
        
        num_missed = 0
        num_false = 0
        
        for word, _, missed, false in self.errors(dictionary, margins):
            w = dictionary.weights[word]
            num_missed += sum(w[i] for i in missed)
            num_false  += sum(w[i] for i in false)
        
        return num_missed, num_false



'''
Created on Mar 7, 2016

@author: mike
'''
import collections

# hyphenation margins
class Margins(collections.namedtuple('_Margins', ['left', 'right'])):

    @classmethod
    def parse(cls, margins):
        parts = margins.split(',')

        if len(parts) != 2:
            raise ValueError('margins format error: expect two numbers delimitered with a comma, e.g. "1,1". Got: %s' % margins)
    
        return Margins(*tuple(int(x) for x in parts))
    
    def __repr__(self):
        return '%s,%s' % self



'''
Created on Mar 7, 2016

@author: mike
'''
import collections
from patgen import EMPTYSET


class Layer:
    
    @property
    def maxchunk(self):
        return self.patlen_range.end

    def __init__(self, patlen_range, selector, inhibiting):
        self.patlen_range = patlen_range
        self.selector = selector
        self.inhibiting = inhibiting
        
        self._data = collections.defaultdict(set)
    
    def __getitem__(self, i):
        return self._data[i]
    
    def get(self, i, defaultval=None):
        return self._data.get(i, defaultval)

    def __len__(self):
        return len(self._data)
    
    def keys(self):
        return self._data.keys()
    
    def items(self):
        return self._data.items()
    
    def values(self):
        return self._data.values()
    
    def update(self, vals):
        self._data.update(vals)
    
    def __repr__(self):
        return '<Layer patlen_range=%r, selector=%r, data=%r>' % (self.patlen_range, self.selector, self._data)
    
    def compute_num_patterns(self):
        return sum(len(val) for val in self.values())

    def train(self, patlen, dictionary, margins):
        patterns = collections.defaultdict(set)
        for position in range(0, patlen+1):
            for ch, num_good, num_bad in dictionary.generate_pattern_statistics(
                                                                            self.inhibiting, 
                                                                            patlen, 
                                                                            position, 
                                                                            margins=margins):
                    if self.selector.select(num_good, num_bad):
                        patterns[ch].add(position)

        self.update(patterns)
        return patterns

    def predict(self, word, margins):
        '''
        Applies a single pattern layer set to the word
        
        Result is the set of indices that patterset "suggested".
        For hyphenation patternsets, these are indices where hyphenation is predicted.
        For inhibiting patternsets, these are indices where hyphenation is inhibited.
        '''
        word = '.' + word + '.'
    
        prediction = set()
    
        for chunklen in range(1, self.maxchunk+1):
            for start in range(0, len(word) - chunklen):
                ch = word[start: start+chunklen]
                allowed = self._data.get(ch, EMPTYSET)
                for index in allowed:
                    if start + index > margins.left and start+index <= len(word) - 1 - margins.right:
                        prediction.add(index + start - 1)  # -1 corrects for the added front padding
    
        return prediction

    def apply_to_dictionary(self, inhibiting, dictionary, margins):
        num_missed = 0
        num_false  = 0
    
        for word, hyphens in dictionary.items():
            predicted = self.predict(word, margins=margins)
            
            missed = dictionary.missed[word]
            false  = dictionary.false[word]
            weights = dictionary.weights[word]
    
            if not inhibiting:
                missed.difference_update(predicted)
                false.update(predicted - hyphens)
            else:
                false.difference_update(predicted)
                missed.update(hyphens & (predicted - missed))
    
            num_missed += sum(weights[i] for i in missed)
            num_false += sum(weights[i] for i in false)
        
        return num_missed, num_false



'''
Created on Mar 7, 2016

@author: mike
'''
import collections
import codecs
from patgen import FALSE_HYPHEN, MISSED_HYPHEN, TRUE_HYPHEN, DIGITS
from patgen.margins import Margins
from patgen.chunker import Chunker


class Dictionary:
    ''' Hyphenation dictionary.
    Holds set of allowed hyphenation positions for each word.
    Additionally, can store hyphenation errors (missed and false) and hyphenation weights
    '''
    
    def __init__(self):
        self._hyphens = collections.OrderedDict()
        self._weights = {}
        self._missed = {}
        self._false = {}

    @property
    def weights(self):
        return self._weights
    
    @property
    def missed(self):
        return self._missed

    @property
    def false(self):
        return self._false

    def __getitem__(self, key):
        return self._hyphens[key]
    
    def __setitems__(self, key, val):
        self._hyphens[key] = val
    
    def keys(self):
        return self._hyphens.keys()
    
    def items(self):
        return self._hyphens.items()
    
    def values(self):
        return self._hyphens.values()
    
    def compute_total_hyphens(self):
        return sum(len(h) for h in self.values())

    def compute_margins(self):
        margin_left = 1000
        margin_right = 1000
        for word, hyphen in self.items():
            if hyphen:
                hmin = min(hyphen)
                hmax = max(hyphen)
                margin_left = min(margin_left, hmin)
                margin_right = min(margin_right, len(word) - hmax)

        return Margins(margin_left, margin_right)

    def make_all_missed(self):
        for word in self.keys():
            self.missed[word].clear()
            self.missed[word].update(self[word])  # all hyphens are initially missing
            self.false[word].clear()

    @classmethod
    def load(cls, filename):

        with codecs.open(filename, 'r', 'utf-8') as f:
            return cls.from_string(f.read())
    
    @classmethod
    def from_string(cls, string):
        dictionary = cls()
        for line in string.split('\n'):
            line = line.strip()
            if not line:
                continue
            if line.startswith('#'):  # comment line
                continue

            text, hyphens, missed, false, weight = parse_dictionary_word(line)
            dictionary._hyphens[text] = hyphens
            dictionary._weights[text] = weight
            dictionary._missed[text] = missed
            dictionary._false[text] = false

        return dictionary
    
    def save(self, filename):

        with codecs.open(filename, 'w', 'utf-8') as f:
            
            for word, hyphens in self._hyphens.items():
                w = self._weights[word]
                
                f.write(format_dictionary_word(word, hyphens, w))
                f.write('\n')

    def generate_pattern_statistics(self, inhibiting, patt_len, hyphen_position, margins):
        '''
        Takes a dictionary of word hyphenations and
        1. Finds all possible patterns of a given length and given hyphen position, and
        2. Computes performance of each pattern - how many times pattern finds "good" hyphen, and how many time it errors
        '''
        chunker = Chunker(patt_len, margins=margins)
        
        good = collections.defaultdict(int)
        bad  = collections.defaultdict(int)
    
        for word in self.keys():
            hyphens = self[word]
            missed  = self.missed[word]
            false   = self.false[word]
            weight  = self.weights[word]
    
            for start, ch in chunker(word, hyphenpos=hyphen_position):
                index = start + hyphen_position - 1
                w     = weight[start + hyphen_position - 1]
                if not inhibiting:
                    if index in missed:
                        good[ch] += w
                    elif index not in hyphens:
                        bad[ch] += w
                else:
                    if index in  false:
                        good[ch] += w
                    elif index in hyphens and index not in missed:
                        bad[ch] += w
        
        ##print('.x', len([x for x in sorted(set(good.keys()) | set(bad.keys())) if x.startswith('.') ]))
        ##print('x.', len([x for x in sorted(set(good.keys()) | set(bad.keys())) if x.endswith('.') ]))
    
        return [(ch, good[ch], bad[ch]) for ch in sorted(set(good.keys()) | set(bad.keys()))]


def parse_dictionary_word(word):

    text = []
    weights = {}
    hyphens = set()
    missed = set()
    false = set()
    
    default_weight = 1
    if word[0] in DIGITS:
        default_weight = int(word[0])

    weights[0] = default_weight

    for c in word:
        if c == TRUE_HYPHEN:
            hyphens.add(len(text))
        elif c == MISSED_HYPHEN:
            hyphens.add(len(text))
            missed.add(len(text))
        elif c == FALSE_HYPHEN:
            false.add(len(text))
        elif c in DIGITS:
            weights[len(text)] = int(c)
        else:
            text.append(c)
            weights[len(text)] = default_weight

    return ''.join(text).lower(), hyphens, missed, false, weights


def format_dictionary_word(word, hyphens, missed=None, false=None, weights=None):
    text = []

    for i in range(len(word) + 1):
        if i > 0:
            text.append(word[i-1])

        if false is not None and i in false:
            text.append(FALSE_HYPHEN)
        elif missed is not None and i in missed:
            text.append(MISSED_HYPHEN)
        elif i in hyphens:
            text.append(TRUE_HYPHEN)

        if weights is not None:
            if i == 0:
                if weights[0] > 1:
                    text.append(str(weights[0]))
            elif weights[i] != weights[0]:
                text.append(str(weights[i]))

    return ''.join(text)



'''
Created on Mar 7, 2016

@author: mike
'''
import collections


class Selector(collections.namedtuple('_Selector', ['good_weight', 'bad_weight', 'threshold'])):

    @classmethod
    def parse(cls, selector):
        parts = selector.split(':')
        
        if len(parts) != 3:
            raise ValueError('selector format error: expect three values delimitered with a senicolon, e.g. "1:2:10". Got: %s' % selector)

        return cls(*tuple(float(x) for x in parts))
    
    def select(self, num_good, num_bad):
        return num_good * self.good_weight - num_bad * self.bad_weight >= self.threshold

    def __repr__(self):
        return '%s:%s:%s' % self



'''
Created on Mar 7, 2016

@author: mike
'''
import collections


class Range(collections.namedtuple('_Range', ['start', 'end'])):
    
    @classmethod
    def parse(cls, rng):
        parts = rng.split('-')
        
        if len(parts) != 2:
            raise ValueError('failed to parse Range: expect two integers delimitered with a dash, e.g. "1-3". Got: %s' % rng)
        
        return cls(*tuple(int(x) for x in parts))
    
    def __repr__(self):
        return '%s-%s' % self



'''
Created on Mar 7, 2016

@author: mike
'''
from patgen.margins import Margins

class Chunker:
    
    def __init__(self, chunklen, margins=Margins(1,1)):
        self.chunklen = chunklen
        self.margins = margins

    def __call__(self, word, hyphenpos):
        '''
        Takes word :word: and generates all chunks of the given length with
        hyphen position :hyphenpos:.
        
        margin_left sets the minimal length of word prefix where hyphenation is not allowed.
            this is the same as TeX's \lefthyphenmin
        margin_right sets the minimal length of word suffix where hyphention is not allowed.
            this is the same as TeX's \righthyphenmin
        
        Example:
            chunker = Chunker(2)
            chunker('mike', 2) will produce this sequence:
            0, ".m"
            1, "mi"
            2, "ik"
        '''
        if hyphenpos > len(word):
            return  # word is too short

        assert 0 <= hyphenpos <= len(word)
    
        word = '.' + word + '.'
        
        start = 0
        end = len(word) - self.chunklen + 1 # last valid offset
        
        start = max(start, self.margins.left+1-hyphenpos)
        end = min(end, len(word)-self.margins.right-hyphenpos)
        
        for i in range(start, end):
            yield i, word[i:i+self.chunklen]



'''
Created on Feb 19, 2016

@author: mike
'''
from __future__ import print_function

import codecs
import re
import argparse

TEMPLATE = r'''

\documentclass{article}
\usepackage{polyglossia}
\usepackage{xltxtra}
\usepackage{testhyphens}
\setdefaultlanguage{churchslavonic}
\newfontfamily\churchslavonicfont[Script=Cyrillic]{PonomarUnicode.otf}

\begin{document}

$BODY$

\begin{checkhyphens}
$CHECKHYPHENS$
\end{checkhyphens}

\end{document}
'''


def main_generate(args):

    words = list(read_words(args.dictionary))
    
    body= '\n'.join('\\showhyphens{%s}' % x for x in words)
    
    checkhyphens = '\n'.join(words)

    text = re.sub(r'\$BODY\$', body, TEMPLATE)
    text = re.sub(r'\$CHECKHYPHENS\$', checkhyphens, text)
    
    with codecs.open(args.output, 'w', 'utf-8') as f:
        f.write(text)


def read_words(fname, strip_hyphen=True):
    
    rex = re.compile(r'\d')
    if strip_hyphen:
        rex = re.compile(r'[\-\d]')

    with codecs.open(fname, 'r', 'utf-8') as f:
        for l in f:
            l = l.strip()
            if not l:
                continue
            if l.startswith('#') or l.startswith('%'):
                continue
            l = rex.sub('', l)
            yield l


def main_extract(args):
    
    with codecs.open(args.output, 'w', 'utf-8') as f:
        for x in generatehyps(args.log):
            f.write(x + '\n')


def generatehyps(fname):
    with codecs.open(fname, 'r', 'utf-8') as f:
        
        expect = False
        for l in f:
            l = l.strip()
            if expect > 0:
                expect -= 1

            if expect == 1:
                expect = 0
                yield l.split()[1]
            else:
                if re.search(r'showhyphens: \*\*\*', l):
                    expect = 4


def main_strip_weights(args):

    words = list(read_words(args.input, strip_hyphen=False))

    with codecs.open(args.output, 'w', 'utf-8') as f:
        for word in words:
            f.write(word + '\n')
    
    return 0


def main():
    parser = argparse.ArgumentParser(description='Utility to compare training dictionary with actual hyphenations as generated by TeX')
    sub = parser.add_subparsers(help='Commands', dest='cmd')
    
    parser_generate = sub.add_parser('generate', help='Generate LaTex file with all dictionary words')
    parser_generate.add_argument('dictionary', help='input dictionary file')
    parser_generate.add_argument('output', help='output LaTeX file')
    
    parser_extract = sub.add_parser('extract', help='Extract actual TeX hyphenations from LaTeX log file')
    parser_extract.add_argument('log', help='input LaTeX log file')
    parser_extract.add_argument('output', help='output hyphenation dictionary as extracted from log')

    parser_strip_weights = sub.add_parser('strip_weights', help='Strips weights (numbers) from dictionary')
    parser_strip_weights.add_argument('input', help='input dictionary file')
    parser_strip_weights.add_argument('output', help='output dictionary with numeric weights stripped off')

    args = parser.parse_args()
    
    if args.cmd == 'generate':
        parser.exit(main_generate(args))
    elif args.cmd == 'extract':
        parser.exit(main_extract(args))
    elif args.cmd == 'strip_weights':
        parser.exit(main_strip_weights(args))
    else:
        parser.error('Missing or unrecognized command')


if __name__ == '__main__':
    main()




'''
Created on Mar 9, 2016

@author: mike
'''
from bisect import bisect_left

class SuffixArray:
    def __init__(self, sarray):
        self._sarray = sarray
    
    @classmethod
    def build(cls, values):
        
        sarray = []
        for value in values:
            for suffix, offset in suffixes_with_offset(value):
                sarray.append( (suffix, value, offset) )

        sarray.sort()
        
        return SuffixArray(sarray)
    
    def superstrings(self, query):
        i = bisect_left(self._sarray, (query, '', 0))

        while i < len(self._sarray):
            entry = self._sarray[i]
            if not entry[0].startswith(query):
                break
            assert entry[1][entry[2]:].startswith(query)
            yield entry[1], entry[2]
            i += 1

def suffixes_with_offset(value):
    for i in range(len(value)-1):
        yield value[i:], i


'''
Created on Feb 18, 2016

@author: mike
'''
from __future__ import print_function

import collections
import codecs
import pickle
import os
from datetime import datetime


EMPTYSET = frozenset()

MISSED_HYPHEN = '.'
FALSE_HYPHEN  = '*'
TRUE_HYPHEN   = '-'
NOT_A_HYPHEN  = ' '

DIGITS = '0123456789'


def stagger_range(start, end):
    middle = start + (end-start) // 2
    left = middle - 1
    right = middle + 1
    
    yield middle
    
    while left >= start or right < end:
        if left >= start:
            yield left
            left -= 1
        
        if right < end:
            yield right
            right += 1



from setuptools import setup
from patgen.version import __version__


setup(
    name='pypatgen',
    version=__version__,
    description='TeX hyphenation pattern generator',

    url='https://github.com/pgmmpk/pypatgen',
    #download_url='https://github.com/pgmmpk/pypatgen/tarball/0.1.6',
    
    author='Mike Kroutikov',
    author_email='pgmmpk@gmail.com',

    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
    ],

    keywords='TeX hyphenation',

    packages=['patgen'],

    install_requires=[],

    entry_points={
        'console_scripts': [
            'pypatgen=patgen.main:main',
            'pypatgen_validate=patgen.validate:main'
        ],
    },
)



'''
Created on Feb 16, 2016

@author: mike
'''
from __future__ import print_function

import os
import codecs
import sys
from patgen.margins import Margins
from patgen.dictionary import Dictionary, format_dictionary_word
from patgen.project import Project
from patgen.selector import Selector
from patgen.range import Range
from patgen.version import __version__


def main_new(args):
    print('Creating new project', args.project, 'from dictionary', args.dictionary)
    
    if os.path.exists(args.project):
        print('Project file already exists! Use different name or delete old project first. File %s' % args.project)
        return -1
    
    if not os.path.exists(args.dictionary):
        print('Dictionary file not found', args.dictionary)
        return -1
    
    dictionary = Dictionary.load(args.dictionary)

    if args.margins is None:
        print('Automatically computing hyphenation margins from dictionary')
        margins = dictionary.compute_margins()
    else:
        margins = Margins.parse(args.margins)

    project = Project(dictionary, margins)
    project.save(args.project)

    return main_show(args)


def main_show(args):
    
    if not os.path.exists(args.project):
        print('Project file not found:', args.project)
        return -1
    
    project = Project.load(args.project)
    
    print('Project file', args.project)
    print('\tcreated:', project.created)
    print('\tlast modified:', project.modified)
    print('\tmargins:', project.margins)
    print('\tdictionary size:', len(project.dictionary.keys()))
    #if project.ignore_weights:
    #    print('\tdictionary weights were ignored (-i flag active)')
    print('\ttotal hyphens: (weighted)', project.total_hyphens)
    print('\ttotal missed : (weighted)', project.missed, percent(project.missed, project.total_hyphens))
    print('\ttotal false  : (weighted)', project.false, percent(project.false, project.total_hyphens))
    print('\tnumber of pattern levels:', len(project.patternset))
    
    for i, layer in enumerate(project.patternset):
        if i & 1 == 0:
            print((i+1), 'HYPHENATING patternset, num patterns:', len(layer))
        else:
            print((i+1), 'INHIBITING patternset, num patterns:', len(layer))
        print('\tTrained with: range %r, selector %r' % (layer.patlen_range, layer.selector))

    print()
    return 0


def main_train(args):
    print('Training project', args.project, 'using range', args.range, 'and selector', args.selector)
    
    project = Project.load(args.project)

    if len(project.patternset) & 1:
        print('Training INHIBINTING pattern layer (level=%s)' % (len(project.patternset)+1))
    else:
        print('Training HYPHENATION pattern layer (level=%s)' % (len(project.patternset)+1))

    patlen_rng = Range.parse(args.range)
    selector = Selector.parse(args.selector)

    print('\tpattern lengths:', patlen_rng)
    print('\tselector:', args.selector)
    
    total_hyphens = project.total_hyphens

    project.train_new_layer(patlen_rng, selector)

    missed, false = project.missed, project.false

    print('Missed (weighted):', missed, percent(missed, total_hyphens))
    print('False (weighted):', false, percent(false, total_hyphens))

    if args.commit:
        project.save(args.project)
        print('...Committed!')
    else:
        print('...Projects NOT changed (use --commit flag to save changes)')

    print()
    return 0


def main_batchtrain(args):
    print('Batch-training of project', args.project, 'using specs from', args.specs)

    with codecs.open(args.specs, 'r', 'utf-8') as f:
        l = {}
        exec(f.read(), {}, l)
        batches = l['SPECS']

    for params in batches:
        args.range = params['range']
        args.selector = params['selector']
        args.commit = True
        
        rc = main_train(args)
        if rc:
            print('Error:', rc)
            return rc
    
    return 0


def main_export(args):
    print('Exporting patterns from', args.project, 'and saving them in TeX format to', args.output)

    if os.path.exists(args.output):
        print('Pattern file already exists! Delete it first, or change the name. Pattern file: %s' % args.output)
        return -1
    
    project = Project.load(args.project)

    num_patterns = 0
    num_exceptions = 0
    
    with codecs.open(args.output, 'w', 'utf-8') as f:
        f.write('\\patterns{\n')
        for patt in project.patternset.pattern_strings():
            f.write(patt + '\n')
            num_patterns += 1
        f.write('}\n')
        f.write('\\hyphenation{\n')
        for word, _, _, _ in project.patternset.errors(project.dictionary, project.margins):
            text = format_dictionary_word(word, project.dictionary[word])
            f.write(text + '\n')
            num_exceptions += 1
        f.write('}\n')
    
    print('Created TeX patterns file', args.output)
    print('Number of patterns:', num_patterns)
    print('Number of exceptions:', num_exceptions)

    print()
    return 0


def main_hyphenate(args):
    print('Hyphenating', args.input, 'into', args.output, 'using project', args.project)
    
    project = Project.load(args.project)

    with codecs.open(args.input or sys.stdin.fileno(), 'r', 'utf-8') as f:
        with codecs.open(args.output or sys.stdout.fileno(), 'w', 'utf-8') as out:

            for word in f:
                word = word.strip()
                if not word:
                    continue
    
                prediction = project.patterns.hyphenate(word, margins=project.margins) 

                s = format_dictionary_word(word, prediction)
                out.write(s + '\n')
    
    print()
    return 0


def main_test(args):
    print('Testing', args.project, 'on dictionary', args.dictionary)
    project = Project.load(args.project)

    dictionary = Dictionary.load(args.dictionary)

    print('Performance of', args.project, 'on', args.dictionary)
    do_test(project, dictionary)
    if args.errors:
        with codecs.open(args.errors, 'w', 'utf-8') as f:
            for word, hyphens, missed, false in project.patternset.errors(dictionary, project.margins):
                f.write(format_dictionary_word(word, hyphens, missed, false) + '\n')

    print()
    return 0


def main_swap(args):
    print('Swapping inhibiting layers between', args.project, 'and', args.project2)

    project = Project.load(args.project)
    project2 = Project.load(args.project2)
    
    if len(project.patternset) != len(project2.patternset):
        raise ValueError('You can only swap layers between projects with same number of layers!')
    
    for i in range(1, len(project.patternset), 2):
        project.patternset[i], project2.patternset[i] = project2.patternset[i], project.patternset[i]

    print('Performance of', args.project)
    project.missed, project.false = do_test(project, project.dictionary)
    print('Performance of', args.project2)
    project2.missed, project2.false = do_test(project2, project2.dictionary)
    
    if args.commit:
        project.save(args.project)
        project2.save(args.project2)
        print('...Committed')
    else:
        print('...Projects NOT changed (use --commit flag to save changes)')
    
    print()
    return 0


def main_compact(args):
    print('Compacting hyphenation patterns for', args.project)
    project = Project.load(args.project)

    before_compact = [layer.compute_num_patterns() for layer in project.patternset]
    
    project.patternset.compact()
    
    after_compact =  [layer.compute_num_patterns() for layer in project.patternset]
    
    print('Result:')
    for level0, (before, after) in enumerate(zip(before_compact, after_compact)):
        print('\tLevel %s: %6d => %6d' % (level0+1, before, after))
    
    if args.commit:
        project.save(args.project)
        print('...Committed')
    else:
        print('...Project NOT changed (use --commit flag to save changes)')
    
    print()
    return 0


def do_test(project, dictionary):

    total_hyphens = dictionary.compute_total_hyphens()

    num_missed, num_false = project.patternset.evaluate(dictionary, project.margins)

    print('Missed (weighted):', num_missed, '(%4.3f%%)' % (num_missed * 100 / (total_hyphens + 0.000001)))
    print('False (weighted):', num_false,   '(%4.3f%%)' % (num_false * 100 / (total_hyphens + 0.000001)))
    
    return num_missed, num_false


def main():
    import argparse
    
    parser = argparse.ArgumentParser(description='Generates TeX hyphenation patterns')
    parser.add_argument('project', help='Name of the project file')

    sub    = parser.add_subparsers(help='Commands', dest='cmd')

    # "new" command
    parser_new = sub.add_parser('new', help='Creates new hyphenation pattern project from dictionary')
    parser_new.add_argument('dictionary', help='Dictionary of hyphenated words (one word per line)')
    parser_new.add_argument('-m', '--margins', help='Hyphenation margins. If not set, will  be computed from the dictionary')
    
    # "show" command
    parser_show = sub.add_parser('show', help='Displays information about current hyphenation project')  # @UnusedVariable

    # "train" command
    parser_train = sub.add_parser('train', help='Trains next level of hyphenation patterns')
    parser_train.add_argument('-s', '--selector', default='1:5:10', help='triplet of numbers that control pattern selection. Format is: "good_weight:bad_weight:threshold"')
    parser_train.add_argument('-r', '--range', default='1,5', help='range of patterns lengths. Default is 1,5')
    parser_train.add_argument('-c', '--commit', default=False, action='store_true', help='If set, project is modified')

    # "batchtrain" command
    parser_batchtrain = sub.add_parser('batchtrain', help='Trains all levels from batch specs')
    parser_batchtrain.add_argument('specs', help='file with batch training parameters specifications')

    # "export" command
    parser_export = sub.add_parser('export', help='Exports project as a set of TeX patterns')
    parser_export.add_argument('output', help='Name of the TeX pattern file to create')

    # "hyphenate" command
    parser_hyphenate = sub.add_parser('hyphenate', help='Hyphenates a list of words')
    parser_hyphenate.add_argument('-i', '--input', default=None, help='Input file with word list - one word per line. If not given, reads stdin.')
    parser_hyphenate.add_argument('-o', '--output', default=None, help='Output file with hyphenated words - one per line. If not given, writes to stdout.')

    # "test" command
    parser_test = sub.add_parser('test', help='Test performance on a dictionary')
    parser_test.add_argument('dictionary', help='File name of a test dictionary')
    parser_test.add_argument('-e', '--errors', help='Optional file to write errors (for error analysis)')

    # "swap" command
    parser_swap = sub.add_parser('swap', help='Swaps odd layers between two projects (advanced)')
    parser_swap.add_argument('project2', help='File name of a second project')
    parser_swap.add_argument('-c', '--commit', default=False, action='store_true', help='If set, swapped projects are saved')

    # "compact" command
    parser_compact = sub.add_parser('compact', help='Removes redundancy from parterns')
    parser_compact.add_argument('-c', '--commit', default=False, action='store_true', help='If set, swapped projects are saved')

    if '-v' in sys.argv or '--version' in sys.argv:
        print('PYPATGEN version:', __version__)
        parser.exit(0)

    args = parser.parse_args()
    if args.cmd == 'new':
        parser.exit(main_new(args))
    elif args.cmd == 'show':
        parser.exit(main_show(args))
    elif args.cmd == 'train':
        parser.exit(main_train(args))
    elif args.cmd == 'export':
        parser.exit(main_export(args))
    elif args.cmd == 'hyphenate':
        parser.exit(main_hyphenate(args))
    elif args.cmd == 'batchtrain':
        parser.exit(main_batchtrain(args))
    elif args.cmd == 'test':
        parser.exit(main_test(args))
    elif args.cmd == 'swap':
        parser.exit(main_swap(args))
    elif args.cmd == 'compact':
        parser.exit(main_compact(args))
    else:
        parser.error('Command required')


def percent(x, base):
    return '%4.2f%%' % (100.0 * x / (base + 0.0000001))

if __name__ == '__main__':
    main()



'''
Created on Mar 7, 2016

@author: mike
'''
from datetime import datetime
from patgen.patternset import PatternSet
import pickle
import os
from patgen.layer import Layer
from patgen import stagger_range


class Project:
    
    VERSION = '2016/03/07'

    def __init__(self, dictionary, margins=None, total_hyphens=None):
        
        self.dictionary = dictionary
        self.dictionary.make_all_missed()

        self.margins = margins or dictionary.compute_margins()
        self.total_hyphens = total_hyphens or dictionary.compute_total_hyphens()
        self.created = self.modified = datetime.now()
        self.patternset = PatternSet()

        # statistics
        self.missed = self.total_hyphens
        self.false = 0

    def __getstate__(self):
        return (
            self.VERSION, 
            self.created, 
            datetime.now(), 
            self.dictionary, 
            self.margins, 
            self.total_hyphens, 
            self.patternset, 
            self.missed, 
            self.false
        )
    
    def __setstate__(self, state):
        (
            version, 
            self.created, 
            self.modified, 
            self.dictionary, 
            self.margins, 
            self.total_hyphens, 
            self.patternset, 
            self.missed, 
            self.false
        ) = state
        
        if version != self.VERSION:
            raise RuntimeError('Incompatible version: %s (expected %s)' % (version, self.VERSION))
    
    @classmethod
    def load(cls, filename):

        if not os.path.exists(filename):
            raise RuntimeError('Project file not found: %s' % filename)
        
        with open(filename, 'rb') as f:
            return pickle.load(f)

    def save(self, filename):
        
        with open(filename, 'wb') as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)

    def train_new_layer(self, patlen_range, selector):

        inhibiting = len(self.patternset) & 1
    
        layer = Layer(patlen_range, selector, inhibiting)
        self.patternset.append(layer)
        
        for patlen in stagger_range(patlen_range.start, patlen_range.end + 1):
            additions = layer.train(patlen, self.dictionary, self.margins)
            print('Selected %s patterns of length %s' % (len(additions), patlen))
    
        # evaluate
        missed, false = layer.apply_to_dictionary(inhibiting, self.dictionary, margins=self.margins)
        self.missed = missed
        self.false = false
        
        return layer
